@@ -1,3 +1,4 @@
+import bamt.utils.GraphUtils
 from bamt import Builders, Nodes
 import random
 import re
@@ -11,10 +12,9 @@ import os
 
 from tqdm import tqdm
 from concurrent.futures import ThreadPoolExecutor
-from bamt.utils import GraphUtils as gru
 from pyvis.network import Network
 
-from typing import Dict, Tuple, List, Callable, Optional, Type, Union
+from typing import Dict, Tuple, List, Callable, Optional, Type, Union, Any, Sequence
 from bamt.Builders import ParamDict
 
 from bamt.log import logger_network
@@ -22,6 +22,7 @@ from bamt.config import config
 
 STORAGE = config.get('NODES', 'models_storage', fallback='models_storage is not defined')
 
+
 class BaseNetwork(object):
     """
     Base class for Bayesian Network
@@ -38,19 +39,20 @@ class BaseNetwork(object):
         self._allowed_dtypes = ['Abstract']
         self.nodes = []
         self.edges = []
-        self.descriptor = {}
+        self.descriptor = {"types": {},
+                           "signs": {}}
         self.distributions = {}
         self.has_logit = False
         self.use_mixture = False
 
-
     @property
     def nodes_names(self) -> List[str]:
         return [node.name for node in self.nodes]
 
     def __getitem__(self, node_name: str) -> Type[Nodes.BaseNode]:
-        index = self.nodes_names.index(node_name)
-        return self.nodes[index]
+        if node_name in self.nodes_names:
+            index = self.nodes_names.index(node_name)
+            return self.nodes[index]
 
     def validate(self, descriptor: Dict[str, Dict[str, str]]) -> bool:
         types = descriptor['types']
@@ -68,7 +70,6 @@ class BaseNetwork(object):
         Function for initializing nodes in Bayesian Network
         descriptor: dict with types and signs of nodes
         """
-        self.descriptor = descriptor
         if not self.validate(descriptor=descriptor):
             if not self.type == 'Hybrid':
                 logger_network.error(
@@ -76,10 +77,11 @@ class BaseNetwork(object):
                 return
             else:
                 logger_network.error(
-                    f"Hybrid BN is not supposed to work with only one type of data. Use DiscreteBN or Continuous BN instead.")
+                    f"Descriptor validation failed due to wrong type of column(s).")
                 return
         elif ['Abstract'] in self._allowed_dtypes:
             return None
+        self.descriptor = descriptor
         # LEVEL 1
         worker_1 = Builders.VerticesDefiner(descriptor)
         self.nodes = worker_1.vertices
@@ -104,41 +106,116 @@ class BaseNetwork(object):
                 f"{self.type} BN does not support {'discrete' if self.type == 'Continuous' else 'continuous'} data")
             return None
         if optimizer == 'HC':
+          
             worker = Builders.HCStructureBuilder(data=data,
                                                  descriptor=self.descriptor,
                                                  scoring_function=scoring_function,
                                                  has_logit=self.has_logit,
                                                  use_mixture=self.use_mixture)
+       
+           
             self.sf_name = scoring_function[0]
+            
             worker.build(data=data, params=params, classifier=classifier)
 
             # update family
             self.nodes = worker.skeleton['V']
             self.edges = worker.skeleton['E']
 
-    def set_nodes(self, nodes: Optional[Dict[str, Type[Nodes.BaseNode]]] = None, **kwargs):
+    def set_nodes(self, nodes: List, info: Optional[Dict] = None):
         """
         additional function to set nodes manually. User should be aware that
         nodes must be a subclass of BaseNode.
         :param nodes dict with name and node (if a lot of nodes should be added)
         """
-        if nodes is None:
-            nodes = dict()
-        nodes.update(kwargs)
-        for column_name, node in nodes.items():
+        if not info and not self.descriptor["types"]:
+            logger_network.error("In case of manual setting nodes user should set map for them as well.")
+            return
+        self.nodes = []
+        for node in nodes:
             try:
-                assert issubclass(node, Nodes.BaseNode)
+                assert issubclass(type(node), Nodes.BaseNode)
+                self.nodes.append(node)
+                continue
             except AssertionError:
                 logger_network.error(f"{node} is not an instance of {Nodes.BaseNode}")
                 continue
             except TypeError:
-                logger_network.error(f"Passed kwarg must be a class. Arg: {node}")
+                logger_network.error(f"TypeError : {node.__class__}")
                 continue
+        if info:
+            self.descriptor = info
 
-            self.nodes.append(node(name=column_name))
-            self.update_descriptor()
+    def set_edges(self, edges: Optional[List[Sequence[str]]] = None):
+        """
+        additional function to set edges manually. User should be aware that
+        nodes must be a subclass of BaseNode.
+        :param edges dict with name and node (if a lot of nodes should be added)
+        """
 
-    def get_params_tree(self, outdir: str):
+        if not self.nodes:
+            logger_network.error("Graph without nodes")
+        self.edges = []
+        for node1, node2 in edges:
+            if isinstance(node1, str) and isinstance(node2, str):
+                if self[node1] and self[node2]:
+                    self.edges.append((node1, node2))
+                else:
+                    logger_network.error(f"Unknown Nodes : [{node1}, {node2}]")
+                    continue
+            else:
+                logger_network.error(f"Unknown node(s) type: [{node1.__class__}, {node2.__class__}]")
+                continue
+        self.update_descriptor()
+
+    def set_structure(self,
+                      info: Optional[Dict] = None,
+                      nodes: Optional[List] = None,
+                      edges: Optional[List[Sequence[str]]] = None,
+                      overwrite: bool = True):
+        """
+        Function to set structure manually
+        info: Descriptor
+        nodes, edges:
+        overwrite: use 2 stage of defining or not
+        """
+        if nodes and (info or (self.descriptor["types"] and self.descriptor["signs"])):
+            self.set_nodes(nodes=nodes, info=info)
+        if edges:
+            self.set_edges(edges=edges)
+            if overwrite:
+                builder = Builders.VerticesDefiner(descriptor=self.descriptor)  # init worker
+                builder.skeleton['V'] = builder.vertices  # 1 stage
+                builder.skeleton['E'] = self.edges
+                builder.get_family()
+                if self.edges:
+                    builder.overwrite_vertex(has_logit=self.has_logit, use_mixture=self.use_mixture)
+                    self.set_nodes(nodes=builder.skeleton['V'])
+                else:
+                    logger_network.error("Empty set of edges")
+
+    def _param_validation(self, params: Dict[str, Any]) -> bool:
+        if all(self[i] for i in params.keys()):
+            for name, info in params.items():
+                try:
+                    self[name].choose(node_info=info, pvals=[])
+                except Exception as ex:
+                    logger_network.error("Validation failed", exc_info=ex)
+                    return False
+            return True
+        else:
+            logger_network.error("Param validation failed due to unknown nodes.")
+            return False
+
+    def set_parameters(self, parameters: Dict):
+        if not self.nodes:
+            logger_network.error("Failed on search of BN's nodes.")
+        # elif self._param_validation(parameters):
+            # pass
+
+        self.distributions = parameters
+
+    def save_params(self, outdir: str):
         """
         Function to save BN params to json file
         outdir: output directory
@@ -149,6 +226,45 @@ class BaseNetwork(object):
             json.dump(self.distributions, out)
         return True
 
+    def save_structure(self, outdir: str):
+        """
+        Function to save BN edges to json file
+        outdir: output directory
+        """
+        if not outdir.endswith('.json'):
+            return None
+        with open(outdir, 'w+') as out:
+            json.dump(self.edges, out)
+        return True
+
+    def save(self, outdir: str):
+        """
+        Function to save the whole BN to json file
+        :param outdir: output directory
+        """
+        if not outdir.endswith('.json'):
+            return None
+        outdict = {
+            'info': self.descriptor,
+            'edges': self.edges,
+            'parameters': self.distributions
+        }
+        with open(outdir, 'w+') as out:
+            json.dump(outdict, out)
+        return True
+
+    def load(self, input_dir: str):
+        """
+        Function to load the whole BN from json file
+        :param input_dir: input directory
+        """
+        with open(input_dir) as f:
+            input_dict = json.load(f)
+
+        self.add_nodes(input_dict['info'])
+        self.set_structure(edges=input_dict['edges'])
+        self.set_parameters(parameters=input_dict['parameters'])
+
     def fit_parameters(self, data: pd.DataFrame, dropna: bool = True):
         """
         Base function for parameters learning
@@ -163,7 +279,7 @@ class BaseNetwork(object):
                     os.makedirs(os.path.join(STORAGE, "0"))
                 elif os.listdir(STORAGE):
                     index = sorted(
-                    [int(id) for id in os.listdir(STORAGE)]
+                        [int(id) for id in os.listdir(STORAGE)]
                     )[-1] + 1
                     os.makedirs(os.path.join(STORAGE, str(index)))
 
@@ -172,12 +288,6 @@ class BaseNetwork(object):
             columns_names = [name for name, t in self.descriptor['types'].items() if t in ['disc_num']]
             data[columns_names] = data.loc[:, columns_names].astype('str')
 
-        # Topology sorting
-        ordered = gru.toporder(self.nodes, self.edges)
-        notOrdered = [node.name for node in self.nodes]
-        mask = [notOrdered.index(name) for name in ordered]
-        self.nodes = [self.nodes[i] for i in mask]
-
         def worker(node):
             return node.fit_parameters(data)
 
@@ -208,44 +318,76 @@ class BaseNetwork(object):
                 print(
                     f"{n.name: <20} | {n.type: <50} | {self.descriptor['types'][n.name]: <10} | {str([self.descriptor['types'][name] for name in n.cont_parents + n.disc_parents]): <50} | {str([name for name in n.cont_parents + n.disc_parents])}")
 
-    def sample(self, n: int, evidence: Optional[Dict[str, Union[str, int, float]]] = None,
-               as_df: bool = True) -> Union[None, pd.DataFrame, List[Dict[str, Union[str, int, float]]]]:
+    def sample(self,
+               n: int,
+               evidence: Optional[Dict[str, Union[str, int, float]]] = None,
+               as_df: bool = True,
+               predict: bool = False,
+               parall_count: int = 1) -> \
+            Union[None, pd.DataFrame, List[Dict[str, Union[str, int, float]]]]:
         """
         Sampling from Bayesian Network
         n: int number of samples
         evidence: values for nodes from user
+        parall_count: number of threads. Defaults to 1.
         """
-        seq = []
+        from joblib import Parallel, delayed
+
         random.seed()
         if not self.distributions.items():
             logger_network.error("Parameter learning wasn't done. Call fit_parameters method")
             return None
-        for n in range(n):
-            output = {}
+        if evidence:
             for node in self.nodes:
+                if (node.type == 'Discrete') & (node.name in evidence.keys()):
+                    if not (isinstance(evidence[node.name], str)):
+                        evidence[node.name] = str(int(evidence[node.name]))
+
+        def wrapper(bn, evdnce):
+            output = {}
+            for node in bn.nodes:
                 parents = node.cont_parents + node.disc_parents
-                if evidence:
-                    if node.name in evidence.keys():
-                        output[node.name] = evidence[node.name]
+                if evdnce:
+                    if node.name in evdnce.keys():
+                        output[node.name] = evdnce[node.name]
                     else:
                         if not parents:
                             pvals = None
                         else:
-                            if self.type == 'Discrete':
+                            if bn.type == 'Discrete':
                                 pvals = [str(output[t]) for t in parents]
                             else:
                                 pvals = [output[t] for t in parents]
-                        output[node.name] = node.choose(self.distributions[node.name], pvals=pvals)
+                        if predict:
+                            output[node.name] = \
+                                node.predict(
+                                    bn.distributions[node.name], pvals=pvals)
+                        else:
+                            output[node.name] = \
+                                node.choose(
+                                    bn.distributions[node.name], pvals=pvals)
+
                 else:
                     if not parents:
                         pvals = None
                     else:
-                        if self.type == 'Discrete':
+                        if bn.type == 'Discrete':
                             pvals = [str(output[t]) for t in parents]
                         else:
                             pvals = [output[t] for t in parents]
-                    output[node.name] = node.choose(self.distributions[node.name], pvals=pvals)
-            seq.append(output)
+                    if predict:
+                        output[node.name] = \
+                            node.predict(
+                                bn.distributions[node.name], pvals=pvals)
+                    else:
+                        output[node.name] = \
+                            node.choose(
+                                bn.distributions[node.name], pvals=pvals)
+            return output
+
+        seq = Parallel(n_jobs=parall_count)(
+            delayed(wrapper)(self, evidence)
+            for i in tqdm(range(n), position=0, leave=True))
 
         if as_df:
             return pd.DataFrame.from_dict(seq, orient='columns')
@@ -265,6 +407,10 @@ class BaseNetwork(object):
         Returns:
             predicted data (dict): dict with column as key and predicted data as value
         """
+        if test.isnull().any().any():
+            logger_network.error("Test data contains NaN values.")
+            return {}
+
         from joblib import Parallel, delayed
 
         def wrapper(bn: HybridBN, test: pd.DataFrame, columns: List[str]):
@@ -275,18 +421,15 @@ class BaseNetwork(object):
                     test_row = dict(test.iloc[i, :])
                     for n, key in enumerate(columns):
                         try:
-                            sample = bn.sample(2000, evidence=test_row)
-                            if bn[key].type.startswith(('Discrete', 'Logit', 'ConditionalLogit',)):
-                                count_stats = sample[key].value_counts()
-                                preds[key].append(count_stats.index[0])
-                            else:
-                                if bn.descriptor['signs'][key] == 'pos':
-                                    sample = sample.loc[sample[key] >= 0]
-                                if sample.shape[0] == 0:
-                                    preds[key].append(np.nan)
+                            sample = bn.sample(1, evidence=test_row, predict=True)
+                            if bn.descriptor['types'][key] == 'cont':
+                                if (bn.descriptor['signs'][key] == 'pos') & (sample.loc[0, key] < 0):
+                                    # preds[key].append(np.nan)
+                                    preds[key].append(0)
                                 else:
-                                    pred = np.mean(sample[key].values)
-                                    preds[key].append(pred)
+                                    preds[key].append(sample.loc[0, key])
+                            else:
+                                preds[key].append(sample.loc[0, key])
                         except Exception as ex:
                             logger_network.error(ex)
                             preds[key].append(np.nan)
@@ -310,8 +453,8 @@ class BaseNetwork(object):
             for n, key in enumerate(columns):
                 preds[key].append(curr_pred[key][0])
 
-        for column in columns:
-            preds[column] = [k for k in preds[column] if not pd.isna(k)]
+        # for column in columns:
+        #     preds[column] = [k for k in preds[column] if not pd.isna(k)]
 
         return preds
 
@@ -387,7 +530,6 @@ class BaseNetwork(object):
 
         network.hrepulsion(node_distance=300, central_gravity=0.5)
 
-        import os
         if not (os.path.exists('visualization_result')):
             os.mkdir("visualization_result")
 
@@ -437,4 +579,5 @@ class HybridBN(BaseNetwork):
     def validate(self, descriptor: Dict[str, Dict[str, str]]) -> bool:
         types = descriptor['types']
         s = set(types.values())
-        return True if ({'cont', 'disc', 'disc_num'} == s) or ({'cont', 'disc'} == s) or ({'cont', 'disc_num'} == s) else False
+        return True if ({'cont', 'disc', 'disc_num'} == s) or ({'cont', 'disc'} == s) or (
+                {'cont', 'disc_num'} == s) else False
