@@ -12,8 +12,8 @@ import os
 from os.path import dirname
 import numpy as np
 
-from external.pyBN.readwrite.read import read_bn
-from external.pyBN.inference.map_exact import map_ve_e
+from bamt.external.pyBN.io.read import read_bn
+from bamt.external.pyBN.inference.map_exact import ve_map
 
 class MapExactTestCase(unittest.TestCase):
 
