@@ -4,8 +4,8 @@ __author__ = """N. Cullen <ncullen.th@dartmouth.edu>"""
 from copy import copy
 import numpy as np
 
-from external.pyBN.classes.factor import Factor
-from external.pyBN.classes.factorization import Factorization
+from bamt.external.pyBN.classes.factor import Factor
+from bamt.external.pyBN.classes.factorization import Factorization
 
 
 
